"""Distribute files from a source to a target location."""
import json
import math
import os
import shutil

from asyncio import sleep
from hashlib import sha512
from pathlib import Path

from .handlers import send_message


async def precalculate(config: dict) -> None:
    """Precalculate the source file hashes.

    :param config: The configuration with the paths to precalculate
    :type config: dict
    """
    if 'paths' in config:
        for path in config['paths']:
            if os.path.exists(path['source']):
                hashes = await calculate_hashes(path['source'])
                if not os.path.exists(os.path.join(path['source'], '.ou-container-content')):
                    os.makedirs(os.path.join(path['source'], '.ou-container-content'), exist_ok=True)
                with open(os.path.join(path['source'], '.ou-container-content', 'hashes.json'), 'w') as out_f:
                    json.dump(hashes, out_f)


async def distribute(config: dict) -> None:
    """Distribute a full configuration.

    :param config: The configuration with the paths to copy
    :type config: dict
    """
    send_message({
        'message': 'Determining files to update. This can take a bit. Please wait...'
    })
    send_message({
        'component': 'files',
        'state': 'active',
        'progress': 0,
    })
    if 'paths' in config:
        updates = []
        for idx, path in enumerate(config['paths']):
            send_message({
                'component': 'files',
                'state': 'active',
                'progress': math.floor(100 / len(config['paths']) * idx),
            })
            updates.extend(await determine_updates(path))
            await sleep(0.1)
        send_message({
            'message': 'Updating files...'
        })
        for idx, update in enumerate(updates):
            if update[0] == 'dir':
                os.makedirs(update[1], exist_ok=True)
            elif update[0] == 'file':
                dirname = os.path.dirname(update[2])
                os.makedirs(dirname, exist_ok=True)
                shutil.copyfile(update[1], update[2])
            if idx % 10 == 0:
                send_message({
                    'component': 'files',
                    'state': 'active',
                    'progress': math.floor(100 / len(updates) * idx),
                })
                await sleep(0)
    send_message({
        'message': 'Your files have been updated.'
    })
    send_message({
        'component': 'files',
        'state': 'complete',
        'progress': 100,
    })


async def determine_updates(path: dict) -> list:
    """Distribute a single path configuration.

    This will copy the file contents from source to target. Depending on the overwrite mode, existing files will or
    will not be overwritten.

    :param path: The path configuration to distribute
    :type path: dict
    """
    updates = []
    if not os.path.exists(os.path.join(path['source'], '.ou-container-content', 'hashes.json')):
        hashes = await calculate_hashes(path['source'])
        if not os.path.exists(os.path.join(path['source'], '.ou-container-content')):
            os.makedirs(os.path.join(path['source'], '.ou-container-content'), exist_ok=True)
        with open(os.path.join(path['source'], '.ou-container-content', 'hashes.json'), 'w') as out_f:
            json.dump(hashes, out_f)
    if os.path.exists(path['source']):
        async for basepath, dirnames, filenames in async_walk(path['source']):
            for dirname in dirnames:
                if dirname != '.ou-container-content':
                    targetpath = os.path.join(path['target'],
                                              os.path.join(basepath, dirname)[len(path['source']) + 1:])
                    if not os.path.exists(targetpath):
                        updates.append(('dir', targetpath))
        with open(os.path.join(path['source'], '.ou-container-content', 'hashes.json')) as in_f:
            source_hashes = json.load(in_f)
        target_hashes = await calculate_hashes(path['target'])
        for filepath, hash in source_hashes.items():
            if filepath not in target_hashes:
                updates.append(('file',
                                os.path.join(path['source'], filepath),
                                os.path.join(path['target'], filepath)))
            elif target_hashes[filepath] != hash:
                if path['overwrite'] == 'always':
                    updates.append(('file',
                                    os.path.join(path['source'], filepath),
                                    os.path.join(path['target'], filepath)))
    return updates


async def calculate_hashes(path: str) -> dict:
    """Calculate the hashes for all files in ``path``.

    :param path: The path for which to calculate the hases
    :type path: str
    :return: The calculated hashes
    :retype: dict
    """
    hashes = {}
    async for basepath, dirnames, filenames in async_walk(path):
        if '.ou-container-content' not in basepath:
            for filename in filenames:
                filepath = os.path.join(basepath, filename)
                try:
                    with open(filepath, 'rb') as in_f:
                        hash = sha512(in_f.read())
                    await sleep(0)
                    hashes[filepath[len(path) + 1:]] = hash.hexdigest()
                except Exception:
                    pass
    return hashes


async def async_walk(path: str) -> None:
    """Asynchronously walk a directory tree.

    This is an asynchronous equivalent to os.walk and is achieved through calling ```await sleep(0)``` after iterating
    through a single directory.

    :param path: The path to walk the directory tree from
    :type path: str
    """
    directory = Path(path)
    if directory.exists():
        dirnames = []
        filenames = []
        for entry in directory.iterdir():
            if entry.is_dir():
                dirnames.append(entry.name)
            else:
                filenames.append(entry.name)
        yield [path, dirnames, filenames]
        await sleep(0)
        for dirname in dirnames:
            async for entry in async_walk(os.path.join(path, dirname)):
                yield entry



"""Handle the startup/shutdown process."""
import tornado

from asyncio import sleep

from .handlers import send_message, completed, messageReceivedEvent
from .distributor import distribute, precalculate
from .scripts import run_startup_scripts, run_shutdown_scripts
from .services import start_services, shutdown_services


async def startup(settings):
    """Run the startup process.

    :param settings: The settings to use for startup
    :type settings: dict
    """
    send_message({'message': 'Container starting up...'})
    await distribute(settings)
    await run_startup_scripts(settings)
    await start_services(settings)
    completed()
    await sleep(0.001)
    await messageReceivedEvent.wait()
    tornado.ioloop.IOLoop.current().stop()


async def shutdown(settings):
    """Run the shutdown process.

    :param settings: The settings to use for shutdown
    :type settings: dict
    """
    await shutdown_services(settings)
    await run_shutdown_scripts(settings)
    await sleep(0.001)
    tornado.ioloop.IOLoop.current().stop()


async def prepare(settings):
    """Run the content preparation process.

    :param settings: The settings to use for shutdown
    :type settings: dict
    """
    await precalculate(settings)
    await sleep(0.001)
    tornado.ioloop.IOLoop.current().stop()



"""Request handlers for the tornado application."""
import json
import tornado.web
import tornado.websocket

from asyncio import Event
from importlib.resources import open_binary
from mimetypes import guess_type


messages = []
complete = False
messageEvent = Event()
messageReceivedEvent = Event()


def send_message(msg: dict):
    """Send a message.

    Notifies all websocket handlers that a new message is to be sent.

    :param msg: The message to send
    :type msg: dict
    """
    messages.append(msg)
    messageEvent.set()
    messageEvent.clear()


def completed():
    """Indicate that the processing is complete."""
    global complete

    complete = True
    messageEvent.set()


class StaticHandler(tornado.web.RequestHandler):
    """Handles sending the static frontend files."""

    def initialize(self, base_path=None):
        """Initialise the :class:`~ou_container_content.handlers.StaticHandler`.

        :param base_path: The base path to prepend to any matched paths
        :type base_path: str
        """
        self._base_path = base_path

    def get(self, path=None):
        """Handle GET requests.

        :param path: The matched path of the file to fetch
        :type path: str
        """
        if path:
            full_path = f'{self._base_path}{path}'
        else:
            full_path = self._base_path
        elements = full_path.split('/')
        with open_binary(f'ou_container_content.frontend{"." if len(elements) > 1 else ""}{".".join(elements[:-1])}',
                         elements[-1]) as in_f:
            mimetype = guess_type(elements[-1])
            if mimetype and mimetype[0]:
                self.set_header('Content-Type', mimetype[0])
            self.write(in_f.read())


class WebsocketHandler(tornado.websocket.WebSocketHandler):
    """Handles sending progress messages via the websocket."""

    async def open(self):
        """Handle opening the websocket connection and sending progress messages."""
        next_idx = 0
        for message in messages:
            await self.write_message(json.dumps(message))
            next_idx = next_idx + 1
        while not complete:
            await messageEvent.wait()
            while next_idx < len(messages):
                await self.write_message(json.dumps(messages[next_idx]))
                next_idx = next_idx + 1
        self.close()
        messageReceivedEvent.set()


async def console_handler():
    """Console handler printing messages to the console."""
    next_idx = 0
    for message in messages:
        if 'message' in message:
            print(message['message'])  # noqa: T001
        next_idx = next_idx + 1
    while not complete:
        await messageEvent.wait()
        while next_idx < len(messages):
            if 'message' in messages[next_idx]:
                print(messages[next_idx]['message'])  # noqa: T001
            next_idx = next_idx + 1
    messageReceivedEvent.set()



"""Configuration validator for the OU Container Content config.yaml."""
from cerberus import Validator
from typing import Union


def null_to_list(value):
    """Coerce None values to an empty list."""
    if value is None:
        return []
    return value


schema = {
    'paths': {
        'type': 'list',
        'required': False,
        'schema': {
            'type': 'dict',
            'schema': {
                'source': {
                    'type': 'string',
                    'required': True,
                    'empty': False
                },
                'target': {
                    'type': 'string',
                    'required': True,
                    'empty': False
                },
                'overwrite': {
                    'type': 'string',
                    'required': True,
                    'allowed': ['always', 'never', 'if-unchanged']
                }
            }
        },
        'coerce': null_to_list
    },
    'scripts': {
        'type': 'dict',
        'required': False,
        'schema': {
            'startup': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'cmd': {
                            'type': 'string',
                            'required': True,
                            'empty': False
                        }
                    }
                },
                'coerce': null_to_list
            },
            'shutdown': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'cmd': {
                            'type': 'string',
                            'required': True,
                            'empty': False
                        }
                    }
                },
                'coerce': null_to_list
            }
        }
    },
    'services': {
        'type': 'list',
        'required': False,
        'schema': {
            'type': 'string',
            'empty': False
        },
        'coerce': null_to_list
    }
}


def validate_settings(settings: dict) -> Union[dict, bool]:
    """Validate the configuration settings against the configuration schema.

    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :return: The validated and normalised settings if they are valid, otherwise ``False``
    :rtype: boolean or dict
    """
    validator = Validator(schema)
    if settings is None:
        return {}
    elif validator.validate(settings):
        return validator.document
    else:
        error_list = []

        def walk_error_tree(err, path):
            if isinstance(err, dict):
                for key, value in err.items():
                    walk_error_tree(value, path + (str(key), ))
            elif isinstance(err, list):
                for sub_err in err:
                    walk_error_tree(sub_err, path)
            else:
                error_list.append(f'{".".join(path)}: {err}')

        walk_error_tree(validator.errors, ())
        return error_list



"""Run scripts."""
import math

from asyncio.subprocess import create_subprocess_exec

from .handlers import send_message


async def run_startup_scripts(settings):
    """Run all startup scripts.

    :param config: The configuration with the scripts to run
    :type config: dict
    """
    send_message({
        'message': 'Running startup scripts...'
    })
    send_message({
        'component': 'scripts',
        'state': 'active',
        'progress': 0,
    })
    if 'scripts' in settings and 'startup' in settings['scripts']:
        for idx, script in enumerate(settings['scripts']['startup']):
            if 'cmd' in script:
                proc = await create_subprocess_exec(*script['cmd'].split(' '))
                await proc.wait()
            send_message({
                'component': 'scripts',
                'state': 'active',
                'progress': math.floor(100 / len(settings['scripts']) * idx),
            })
    send_message({
        'component': 'scripts',
        'state': 'complete',
        'progress': 100,
    })


async def run_shutdown_scripts(settings):
    """Run all shutdown scripts.

    :param config: The configuration with the scripts to run
    :type config: dict
    """
    if 'scripts' in settings and 'shutdown' in settings['scripts']:
        for script in settings['scripts']['shutdown']:
            if 'cmd' in script:
                proc = await create_subprocess_exec(*script['cmd'].split(' '))
                await proc.wait()



"""Start and stop services."""
import math

from asyncio.subprocess import create_subprocess_exec

from .handlers import send_message


async def start_services(settings):
    """Start all services.

    :param config: The configuration with the services to start
    :type config: dict
    """
    send_message({
        'message': 'Starting services...'
    })
    send_message({
        'component': 'services',
        'state': 'active',
        'progress': 0,
    })
    if 'services' in settings:
        for idx, service in enumerate(settings['services']):
            proc = await create_subprocess_exec('sudo', 'service', service, 'start')
            await proc.wait()
            send_message({
                'component': 'services',
                'state': 'active',
                'progress': math.floor(100 / len(settings['scripts']) * idx),
            })
    send_message({
        'component': 'services',
        'state': 'complete',
        'progress': 100,
    })


async def shutdown_services(settings):
    """Shutdown all services.

    :param config: The configuration with the services to stop
    :type config: dict
    """
    if 'services' in settings:
        for service in settings['services']:
            proc = await create_subprocess_exec('sudo', 'service', service, 'stop')
            await proc.wait()



"""The OU Container Content distribution application commandline interface."""
import click
import os
import tornado.ioloop
import tornado.web

from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from .handlers import WebsocketHandler, StaticHandler, console_handler
from . import process
from .validator import validate_settings


def make_app():
    """Build the tornado web application."""
    return tornado.web.Application([
        (r".*websocket", WebsocketHandler),
        (r".*build/(.*)", StaticHandler, {'base_path': 'build/'}),
        (r".*(global.css)", StaticHandler, {'base_path': ''}),
        (r".*(ou-favicon-[0-9]+.png)", StaticHandler, {'base_path': ''}),
        (r".*", StaticHandler, {'base_path': 'index.html'}),
    ])


@click.group()
@click.option('-c', '--config',
              type=click.File(),
              default='/etc/module-content/config.yaml',
              help='The configuration file to use')
@click.pass_context
def main(ctx: click.Context, config: click.File):
    """OU Container Content distribution."""
    settings = load(config, Loader=Loader)
    settings = validate_settings(settings)
    if isinstance(settings, dict):
        ctx.obj = {'settings': settings}
    else:
        click.echo(click.style('There are errors in your configuration settings:', fg='red'), err=True)
        click.echo(err=True)

        for error in settings:
            click.echo(error, err=True)

        raise click.Abort()


@click.command()
@click.pass_context
def startup(ctx: click.Context):
    """Run the startup process."""
    app = make_app()
    app.listen(8888)
    if 'JUPYTERHUB_API_TOKEN' not in os.environ:
        tornado.ioloop.IOLoop.current().add_callback(console_handler)
    tornado.ioloop.IOLoop.current().add_callback(process.startup, ctx.obj['settings'])
    tornado.ioloop.IOLoop.current().start()


@click.command()
@click.pass_context
def shutdown(ctx: click.Context):
    """Run the shutdown process."""
    tornado.ioloop.IOLoop.current().add_callback(process.shutdown, ctx.obj['settings'])
    tornado.ioloop.IOLoop.current().start()


@click.command()
@click.pass_context
def prepare(ctx: click.Context):
    """Run the distribution preparation process."""
    tornado.ioloop.IOLoop.current().add_callback(process.prepare, ctx.obj['settings'])
    tornado.ioloop.IOLoop.current().start()


main.add_command(startup)
main.add_command(shutdown)
main.add_command(prepare)

if __name__ == '__main__':
    main()
